{{
Next 32KB EEPROM Programmer.
}}

con
  _clkmode      = hw#_clkmode
  _xinfreq      = hw#_xinfreq

  BUFFLEN       = ee#PLEN*4

obj
  hw : "properties"
  ee : "PasmEEPROM"
  sx : "FullDuplexSingleton"

dat org 0
loadfile
  file "jem.bin" '' this defines the file to load into the upper 32K of EEPROM
loadfile_end
loadfile_len  long @loadfile_end-@loadfile+1

pub main
  sx.start(hw#UCRXP, hw#UCTXP, hw#UCMODE, hw#UCBAUD)
  'waitcnt(clkfreq+cnt)
  waitcnt(clkfreq*hw#UCWAIT/2+cnt)
  ee.start(hw#BEESCLP,hw#BEESDAP)

  sx.str(string($d,$d,"Read from Next 32K EEPROM",$d))
  dump(loadfile_len,$8000)
  repeat

  sx.str(string($d,$d,"Programming Next 32K EEPROM with "))
  sx.dec(loadfile_len)
  sx.str(string(" Bytes ..."))

  checkEEalias

  program(@loadfile,loadfile_len,$8000)
  verify(@loadfile,loadfile_len,$8000)
  sx.str(string(" verified. Operation Complete!",$d,$d))

  repeat
    waitcnt(clkfreq+cnt)

pub checkEEalias  | j, n, len

  len := ee#PLEN*2

  ee.readBuffer(ee#EEID,0,len,@buff1)
  ee.readBuffer(ee#EEID,$8000,len,@buff2)

  if compareLongs(@buff1,@buff2,len)
    sx.str(string($d,"Data in EEPROM at address 0 and $8000 are identical.",$d))
    sx.str(string("It is possible that your EEPROM is only 32K.",$d))
    sx.str(string("Would you like to continue programming EEPROM? "))
    if sx.rx <> "y"
      repeat

pub readShow(addr, len) | j, n
  ee.readBuffer(ee#EEID,addr,len,@buff1)
  showbytes(@buff1,len)

pub show(addr, len) | j, n
  repeat j from 0 to len-1 step 4
    if j // 16 == 0
      sx.tx($d)
      sx.tx(" ")
      sx.hex(addr+j,4)
      sx.tx(":")
    sx.tx(" ")
    sx.hex(long[addr+j],8)

pub showbytes(addr, len) | j, n
  repeat j from 0 to len-1
    if j // 16 == 0
      sx.tx($d)
      sx.tx(" ")
      sx.hex(addr+j,4)
      sx.tx(":")
    sx.tx(" ")
    sx.hex(byte[addr+j],2)

pub showEEbytes(addr, len, eeadr) | j
  repeat j from 0 to len-1
    if j // 16 == 0
      sx.tx($d)
      sx.tx(" ")
      sx.hex(eeadr+j,4)
      sx.tx(":")
    sx.tx(" ")
    sx.hex(byte[addr+j],2)

pub compareLongs(b1, b2, len) | n
  repeat n from 0 to len-1 step 4
    if(long[b1+n] <> long[b2+n])
{
      sx.hex(b1+n,4)
      sx.tx(" ")
      sx.hex(long[b1+n],8)
      sx.tx(" ")
      sx.hex(b2+n,4)
      sx.tx(" ")
      sx.hex(long[b2+n],8)
      sx.tx($d)
'}
      return 0
  return 1

pub compareBytes(b1, b2, len) | n
  repeat n from 0 to len-1
    if(byte[b1+n] <> byte[b2+n])
      sx.hex(b1+n,4)
      sx.tx(" ")
      sx.hex(byte[b1+n],2)
      sx.tx(" ")
      sx.hex(b2+n,4)
      sx.tx(" ")
      sx.hex(byte[b2+n],2)
      sx.tx($d)
      return 0
  return 1

pub dump(len, eeaddr) | n, num

  repeat n from 0 to len step ee#PLEN

    num := ee#PLEN
    if (len-n) < num
      num := len-n
    ee.readBuffer(ee#EEID,eeaddr,num,@buff1)
    showEEBytes(@buff1, num, eeaddr)
    eeaddr += num

pub program(addr, len, eeaddr) | n, num, div

  div := 20
  repeat n from 0 to len step ee#PLEN

    num := ee#PLEN
    if (len-n) < num
      num := len-n
    sx.str(string($d,"Writing... "))
    sx.hex(eeaddr, 6)
    showbytes(addr+n,num)
    ee.writeBuffer(ee#EEID,eeaddr,num,addr+n)
    waitcnt(clkfreq/div+cnt)
    sx.str(string($d,"Reading... "))
    readShow(eeaddr, num)
    waitcnt(clkfreq/div+cnt)
    eeaddr += num

pub verify(addr, len, eeaddr) | n, num

  repeat n from 0 to len step ee#PLEN

    num := ee#PLEN
    if (len-n) < num
      num := len-n
    ee.readBuffer(ee#EEID,eeaddr,num,@buff1)
    ifnot (compareBytes(addr+n,@buff1,num))
      sx.str(string($d,"Verify Error."))
      sx.str(string($d,"Expected..."))
      showbytes(addr+n,num)
      sx.str(string($d,"Received..."))
      showBytes(@buff1+n, num)
      repeat
    eeaddr += num

dat
  buff1 byte  0 [BUFFLEN]
  buff2 byte  0 [BUFFLEN]
