{{
EEPROM.spin - A Serial EEPROM read/write buffer interface object.
Copyright (c) 2009 by John Steven Denson (jazzed)
See MIT license terms below.
}}

con
  'PLEN          = 32  ' standard EEPROM PAGE LENGTH
  PLEN          = 128  ' AT24C512 EEPROM PAGE LENGTH

{ Test Code
con
  _clkmode = xtal1 + pll16x
  _xinfreq = 5_000_000

obj
  dis: "FullDuplexSingleton"
    
var
  byte  mybuffer[$2000]

dat
  testbuff     byte $11,$22,$44,$88,$55,$55,$55,$55
    
pub main | n, j, mlen

  start(28,29)
  'dis.start(20)
  dis.start(31,30,0,19200)
  waitcnt(clkfreq+cnt)
  
  n := $8000
  'writeBuffer(EEID,n,8,@testbuff)
  mlen := 4
  repeat
    readBytes(EEID,n,mlen,@mybuffer)
    repeat j from 0 to mlen-1
      dis.hex(byte[@mybuffer+j],2)
      dis.tx(" ")
    dis.tx($d)

  mlen := PLEN
  repeat
    dis.out(0)
    readBuffer(EEID,n,mlen,@mybuffer)
'{
    repeat j from 0 to mlen-1 step 4
      if j // 16 == 0
        dis.out($d)
        dis.out(" ")
        dis.hex(n+j,5)
        dis.out(":")
      dis.out(" ")
      dis.hex(long[@mybuffer+j],8)
'}
{
    n += mlen
    n &= $ffff
'}
'    repeat
    waitcnt(clkfreq/2+cnt)

'}

con
  EEID          = $A0
  RBIT          = 1   ' devid byte read bit
  WRBUF         = 1   ' if odd number of bits, write
  RDBUF         = 3   ' if even number of bits, read
  RDBYTES       = 4   ' if bit2 set, read small number of bytes

dat
  command      long  0          ' use for pins on startup
  device       long  EEID       ' default device number
  address      long  0          ' read address
  count        long  0          ' bytes to read
  buffer       long  0          ' buffer
  earlyack     long  0          ' early ack for cache miss

dat ' just one singleton instance at a time please
  cog          long  0

        
pub start(scl_pin, sda_pin)
{{
This is for the main Propeller eeprom.
Other devices could be used by the BigSpin so we let the device define the interface.
It is done this way so we don't have to constantly hack the BigSpin start code.
}}
  command := scl_pin | sda_pin << 16 
  cog := cognew(@entry, @command) + 1
  repeat until command == 0
  
pub stop
  if cog > 0
    cogstop(cog~ -1)

pub cmdptr
  return @command
  
pub pasm
  return @entry

pub size
  return @eof-@entry
  
pub readBuffer(mdev, maddr, mlen, buffptr)
  device  := mdev
  address := maddr
  count   := mlen
  buffer  := buffptr
  command := RDBUF
  repeat until command == 0

pub writeBuffer(mdev, maddr, mlen, buffptr)
  device  := mdev
  address := maddr
  count   := mlen
  buffer  := buffptr
  command := WRBUF
  repeat until command == 0

pub readBytes(mdev, maddr, mlen, buffptr)
  device  := mdev
  address := maddr
  count   := mlen
  buffer  := buffptr
  command := RDBYTES
  repeat until command == 0


con
  eeBufferLen = $20

var
  byte  eeBuffer[eeBufferLen]

pub readByte(maddr)
  readBuffer(EEID, maddr, 1, @eeBuffer)
  return eeBuffer[0]

pub writeByte(maddr,value)
  eeBuffer[0] := value
  writeBuffer(EEID, maddr, 1, @eeBuffer)

dat
  entry        org 0

  jmp          #init

scl            long 0   ' <-- res 1
sda            long 0

tp             long 0   ' temp pointer
dp             long 0   ' device number pointer
ap             long 0   ' addr pointer
bp             long 0   ' buff pointer
sp             long 0   ' size/count pointer

t0             long 0   ' temp vars
t1             long 0

cmd            long 0   ' command word
devs           long 0   ' device select
addr           long 0   ' address
bytes          long 0   ' size in bytes
psiz           long 0   ' parameter size in bytes
buff           long 0   ' buffer pointer
ueack          long 0   ' user early ack
ack            long 0   ' ack bit
val            long 0   ' accumulated value

zero           long 0

'-------------------------------------------------------------------------------
' start point
'
init

  mov          tp,  par ' long 0 word 0
  rdword       t0,  tp
  mov          scl, #1
  shl          scl, t0  ' scl pin bit    

  add          tp, #2   ' word 1
  rdword       t0,  tp
  mov          sda, #1
  shl          sda, t0  ' sda pin bit
    
  add          tp,  #2  ' long 1
  mov          dp,  tp  ' device number pointer
  add          tp,  #4  ' long 2
  mov          ap,  tp  ' address pointer
  add          tp,  #4  ' long 3
  mov          sp,  tp  ' size/count pointer
  add          tp,  #4  ' long 4
  mov          bp,  tp  ' buffer pointer
  add          tp,  #4  ' long 5
  mov          ueack,tp ' early ack pointer

:initlp0
  mov          outa,    scl             ' scl 1
  mov          dira,    scl             ' scl output
  andn         dira,    sda             ' sda output

  jmp          #cmdDone ' clear par so we don't have command pending at startup

'-------------------------------------------------------------------------------
' command loop
'
cmdDone
  wrlong       zero,    par
doCmd
  rdlong       cmd,     par wz
  if_z  jmp    #doCmd

  rdlong       bytes,   sp              ' get number of bytes to read
  rdlong       devs,    dp              ' get device select
  rdlong       addr,    ap              ' get address to read
  rdlong       tp,      bp              ' get buffer address for read

  cmp          cmd,     #RDBYTES wc,wz  ' if bit 2 set, use early ack
  if_e  jmp    #cmdReadBytes

  test         cmd,     #RDBUF wc       ' if c, odd bits for write
  if_nc jmp    #cmdReadBuf              ' else even bits for read
  jmp          #cmdWriteBuf

  jmp          #cmdDone                 ' this should never run
'
'-------------------------------------------------------------------------------


{{
'-------------------------------------------------------------------------------
' read buffer from eeprom
' @param       devs - device number
' @param       addr - read address
' @param       bytes - number of bytes to read
' @param       tp   - address of buffer for read
'
}}
cmdReadBuf
:loop
  cmp          bytes,   #PLEN wc,wz
  if_be mov    psiz,    bytes
  if_a  mov    psiz,    #PLEN
  call         #readPage
  add          addr,    psiz
  sub          bytes,   psiz wc,wz
  if_a  jmp    #:loop
  jmp          #cmdDone

{{
'-------------------------------------------------------------------------------
' write buffer to eeprom
' @param       *dp - device number
' @param       *ap - write address
' @param       *sp - number of bytes to write
' @param       *bp - address of buffer to write
'
}}
cmdWriteBuf
:loop
  cmp          bytes,   #PLEN wc,wz
  if_be mov    psiz,    bytes
  if_a  mov    psiz,    #PLEN
  call         #writePage
  add          addr,    psiz
  sub          bytes,   psiz wc,wz
  if_a  jmp    #:loop
  jmp          #cmdDone


'-------------------------------------------------------------------------------
' read eeprom page
' @param       *dp - device number
' @param       *ap - read address
' @param       *bp - address of buffer for read
' @param       psiz - number of bytes to read
'
readPage
  call         #sendAddr
  or           devs,    #RBIT   ' send read command
  call         #sendCurrent
  mov          t1,      psiz    ' read psiz bytes
  sub          t1,      #1      ' first psiz-1 are NAK
  tjz          t1,      #:last
  andn         ack,     sda     ' not last byte     
:loop
  call         #recVal          ' read page data
  wrbyte       val,     tp
  add          tp,      #1
  djnz         t1,      #:loop
:last  
  or           ack,     sda     ' last byte has ACK     
  call         #recVal
  wrbyte       val,     tp
  add          tp,      #1      ' increment for next loop too
  call         #sendEnd
readPage_ret   ret

{{
'-------------------------------------------------------------------------------
' read buffer from eeprom with early ack after 1 byte read
' @param       devs - device number
' @param       addr - read address
' @param       bytes - number of bytes to read
' @param       tp   - address of buffer for read
'
}}
cmdReadBytes
  'mov          psiz,    bytes
  'call         #readPage
  call         #sendAddr
  or           devs,    #RBIT   ' send read command
  call         #sendCurrent
  mov          t1,      bytes   ' read psiz bytes
  'sub          t1,      #1      ' first psiz-1 are NAK
  tjz          t1,      #:last
  andn         ack,     sda     ' not last byte
:loop
  call         #recVal          ' read page data
  wrbyte       val,     tp
  add          tp,      #1
  djnz         t1,      #:loop
:last
  or           ack,     sda     ' last byte has ACK
  call         #recVal
  wrbyte       val,     tp
  add          tp,      #1      ' increment for next loop too
  call         #sendEnd
  jmp          #cmdDone

'-------------------------------------------------------------------------------
' read current byte
' @param       *dp - device number
' @param       *ap - read address
' @param       *sp - number of bytes to read
' @param       *bp - address of buffer for read
'
readCurrent
  or           devs,    #RBIT   ' read
  call         #sendCurrent
  or           ack,     sda     ' last byte     
  call         #recVal
  wrbyte       val,     tp
  add          tp,      #1
  call         #sendEnd
readCurrent_ret   ret

'-------------------------------------------------------------------------------
' write eeprom page
' @param       *dp - device number
' @param       *ap - write address
' @param       *bp - address of buffer for write
' @param       psiz - number of bytes to read
'
writePage
  call         #sendAddr        ' send page address
  mov          t1,      psiz
:loop
  rdbyte       val,     tp
  add          tp,      #1
  call         #sendVal         ' send data
  djnz         t1,      #:loop
  call         #writeEndPoll
writePage_ret  ret

'-------------------------------------------------------------------------------
' send end and poll for write complete
'
writeEndPoll
  call         #sendEnd         ' send stop
  call         #sclDelay
  call         #sendCurrent     ' now start write polling
  or           outa,    scl
  andn         dira,    sda     ' looking for ACK
:loop
  test         sda,     ina wz  ' is ack 0 ?
  if_nz jmp    #:loop           ' if ack 1, repeat
writeEndPoll_ret  ret  

'-------------------------------------------------------------------------------
' write current eeprom byte
' @param       *dp - device number
' @param       *ap - write address
' @param       *sp - number of bytes to write
' @param       *bp - address of buffer for write
'
writeCurrent
  andn         devs,    #RBIT
  call         #sendStart
  rdbyte       val,     tp
  add          tp,      #1
  call         #sendVal
writeCurrent_ret  ret

'-------------------------------------------------------------------------------
' send eeprom address
' @param       devs - device number
' @param       addr - read address

'
sendAddr
  call         #sendStart
  mov          val,     addr            ' %00000000_00000111_11111111_11111111 ->
  ror          val,     #16             ' %11111111_11111111_00000000_00000111
  shl          val,     #1              ' %11111111_11111110_00000000_00001110 ->
  andn         devs,    #RBIT           ' %10100000 send addr/write needs bit 0 low 
  or           val,     devs            ' %11111111_11111110_00000000_10101110
  call         #sendVal                 ' send device id and upper address bits
  mov          val,     addr            ' %00000000_00000111_11111111_11111111 ->
  ror          val,     #8              ' %11111111_00000000_00000111_11111111
  call         #sendVal                 ' send middle address bits
  mov          val,     addr            ' %00000000_00000111_11111111_11111111
  call         #sendVal                 ' send lower address bits
sendAddr_ret   ret

'-------------------------------------------------------------------------------
' send current is used to initiate a transaction after sendAddr
' @param       devs - device number with bit 0 set for read or clear for write
' @param       addr - read address
'
sendCurrent
  call         #sendStart
  mov          val,     addr            ' %00000000_00000111_11111111_11111111 ->
  ror          val,     #16             ' %11111111_11111111_00000000_00000111
  shl          val,     #1              ' %11111111_11111110_00000000_00001110 ->
  or           val,     devs            ' %11111111_11111110_00000000_1010111r
  call         #sendVal                 ' send device id and upper address bits
sendCurrent_ret  ret


'-------------------------------------------------------------------------------
' A start signal is when SDA goes 1 -> 0 while SCL is 1
'
sendStart
  or           outa,    scl             ' scl high
  or           dira,    scl             ' scl output ... but it is always output right?
  or           outa,    sda             ' sda high
  or           dira,    sda             ' sda output
  call         #sclDelayD2              ' device detects high
  andn         outa,    sda             ' sda low for start condition 
  call         #sclDelayD2              ' device detects start
  andn         outa,    scl             ' finish with scl low 
sendStart_ret  ret
  
'-------------------------------------------------------------------------------
' An end signal is when SDA goes 0 -> 1 while SCL is 1
'
sendEnd
  andn         outa,    sda             ' sda high
  or           outa,    scl             ' scl high
  call         #sclDelayD2
  or           outa,    sda             ' sda high
  call         #sclDelayD2              ' device detects stop
  andn         dira,    sda             
  andn         dira,    scl             ' leave sda & scl input so other devices can use the bus
sendEnd_ret    ret

'-------------------------------------------------------------------------------
' scl has limits on how fast it can go. use this routine to throttle scl
'
sclDelay
  call         #sclDelayD2
  call         #sclDelayD2
sclDelay_ret   ret

sclDelayD2
  nop
'  nop
sclDelayD2_ret ret

'-------------------------------------------------------------------------------
' send 8 bits of data
' @param ack
' @param val
'
sendVal
  mov          ack,     #0              ' %00000000_00000000_00000000_11111111 ->
  ror          val,     #8              ' %11111111_00000000_00000000_00000000
  andn         outa,    scl             ' start with scl 0
  mov          t0,      #8              ' Output val on sda
:loop
  shl          val,     #1 wc           ' get next bit state into carry              
  muxc         outa,    sda             ' set bit depending on carry
  call         #sclDelay
  or           outa,    scl             ' toggle clock
  call         #sclDelay
  andn         outa,    scl
  call         #sclDelayD2
  djnz         t0,      #:loop
:findack  
  andn         dira,    sda             ' looking for ACK
  mov          ack,     ina             ' get ack
  and          ack,     sda wz          ' keep only ack bit
  or           outa,    scl             ' scl high
  call         #sclDelay
  andn         outa,    scl             ' scl low
  andn         outa,    sda             ' sda low
  or           dira,    sda             ' sda back to output
sendVal_ret    ret


'-------------------------------------------------------------------------------
' receive 8 bits of data
' @param ack
' @param val
'
recVal
  mov          val,     #0
  andn         dira,    sda             ' sda input
  andn         outa,    scl             ' clock starts at 0
  call         #sclDelay
  mov          t0,      #8              ' input 8 bits of val from sda
:loop
  or           outa,    scl             ' toggle clock
  call         #sclDelayD2
  shl          val,     #1
  test         sda,     ina wz          ' get bit in z
  muxnz        val,     #1
  andn         outa,    scl             ' toggle clock
  call         #sclDelay
  djnz         t0,      #:loop
:sendack
  or           dira,    sda             ' set sda to out so device can interpret ACK
  or           outa,    ack             ' ack is always on sda, but not always set
  call         #sclDelayD2
  or           outa,    scl        
  call         #sclDelay
  andn         outa,    scl             ' final clock pulse
  andn         outa,    sda        
recVal_ret     ret

'pad            long 0 [($1fc-pad)]      ' pad out to 2KB ... the HUB memory startup area
                                         ' provides a buffer for loading COGS from eeprom 
eof            long $55555555           
  fit          $1ff

{{
 TERMS OF USE: MIT License                                                           

 Permission is hereby granted, free of charge, to any person obtaining a copy
 of this software and associated documentation files (the "Software"), to deal
 in the Software without restriction, including without limitation the rights
 to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
 copies of the Software, and to permit persons to whom the Software is
 furnished to do so, subject to the following conditions:

 The above copyright notice and this permission notice shall be included in
 all copies or substantial portions of the Software.

 THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
 IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
 FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
 AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
 LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE,ARISING FROM,
 OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN
 THE SOFTWARE.
}}